/**
 * Copyright (c) 2017 Lucky Byte, Inc.
 */
const debug = require('debug')('unipay');
const crypto = require('crypto');
const moment = require('moment');
const winston = require('winston');
const utils = require('../lib/utils');
const post = require('../lib/post');

/**
 * 对账文件下载
 *
 * 参数:
 *     dest_url: 银联对账文件下载请求地址
 *     params: 一个 JSON 对象，包含下列的 key:
 *      * merno: 商户代码
 *      * cert_sn: 签名证书 Serial Number
 *      * sign_pem: 签名证书，PEM
 *      * verify_cer: 验证响应报文签名证书
 *      * mmdd: 银联和入网机构间的交易结算日期
 *
 * 返回:
 *     JSON，包含 2 个字段:
 *          reqt: 请求报文, JSON
 *          resp: 响应报文, JSON, 通过报文中 respCode 确定交易是否成功
 */
const statement = async (dest_url, params) => {
    if (!dest_url || !params) {
        throw new Error('银联对账文件下载请求参数无效，用法 statement(dest_url, {...})');
    }
    const required_keys = [
        'merno', 'cert_sn', 'sign_pem', 'verify_cer', 'mmdd'
    ]
    for (let i = 0; i < required_keys.length; i++) {
        const key = required_keys[i];
        if (!params[key]) {
            throw new Error(`银联对账文件下载请求参数缺少[${key}]`);
        }
    }
    let reqt_json = {
        version: '5.0.0',
        encoding: 'UTF-8',
        signMethod: '01',
        txnType: '76',
        txnSubType: '01',
        bizType: '000000',
        accessType: '0',
        txnTime: moment().format('YYYYMMDDHHmmss'),
        certId: params.cert_sn,
        merId: params.merno,
        fileType: '00',
        settleDate: params.mmdd,
    }
    if (params.version) {
        if (params.version != '5.0.0' && params.version != '5.1.0') {
            throw new Error('参数 version 无效，必须是 5.0.0 或 5.1.0');
        }
        reqt_json.version = params.version;
    }
    reqt_json.signature = utils.gen_reqt_sign(reqt_json, params.sign_pem);

    winston.info('银联对账文件下载请求报文：', JSON.stringify(reqt_json, null, 2));

    const resp_data = await post(dest_url, reqt_json);
    const resp_json = utils.parse_resp_data(resp_data);

    winston.info('银联对账文件下载响应报文：', JSON.stringify(resp_json, null, 2));

    if (!utils.verify_resp_sign(resp_json, params.verify_cer)) {
        throw new Error('验证对账文件下载响应报文签名失败');
    }
    winston.info('验证对账文件下载响应报文签名通过.');

    return { reqt: reqt_json, resp: resp_json }
}


module.exports = {
    statement: statement,
}
