/**
 * Copyright (c) 2017 Lucky Byte, Inc.
 */
const debug = require('debug')('unipay');
const crypto = require('crypto');
const moment = require('moment');
const winston = require('winston');
const utils = require('../lib/utils');
const post = require('../lib/post');

/**
 * 被扫扣款
 *
 * 参数:
 *     dest_url: 银联被扫扣款请求地址
 *     params: 一个 JSON 对象，包含下列的 key:
 *      * merno: 商户代码
 *      * cert_sn: 签名证书 Serial Number
 *      * sign_pem: 签名证书，PEM
 *      * verify_cer: 验证响应报文签名证书
 *      * order_id: 商户订单号
 *      * amount: 金额
 *      * qrno: C2B码
 *      * back_url: 后台通知地址
 *      * termid: 终端编号(可选)
 *
 * 返回:
 *     JSON，包含 2 个字段:
 *          reqt: 请求报文, JSON
 *          resp: 响应报文, JSON, 通过报文中 respCode 确定交易是否成功
 */
const consume = async (dest_url, params) => {
    if (!dest_url || !params) {
        throw new Error('银联被扫扣款请求参数无效，用法 consume(dest_url, {...})');
    }
    const required_keys = [
        'merno', 'cert_sn', 'sign_pem', 'verify_cer', 'order_id', 'amount', 'qrno',
        'back_url'
    ]
    for (let i = 0; i < required_keys.length; i++) {
        const key = required_keys[i];
        if (!params[key]) {
            throw new Error(`银联被扫扣款请求参数缺少[${key}]`);
        }
    }
    let reqt_json = {
        version: '5.0.0',
        encoding: 'UTF-8',
        signMethod: '01',
        txnType: '01',
        txnSubType: '06',
        bizType: '000000',
        accessType: '0',
        channelType: '08',
        txnTime: moment().format('YYYYMMDDHHmmss'),
        certId: params.cert_sn,
        merId: params.merno,
        orderId: params.order_id,
        qrNo: params.qrno,
        currencyCode: 156,
        txnAmt: params.amount,
        termId: params.termid || '80000001',
        backUrl: params.back_url,
    }
    if (params.version) {
        if (params.version != '5.0.0' && params.version != '5.1.0') {
            throw new Error('参数 version 无效，必须是 5.0.0 或 5.1.0');
        }
        reqt_json.version = params.version;
    }
    reqt_json.signature = utils.gen_reqt_sign(reqt_json, params.sign_pem);

    winston.info('银联被扫扣款请求报文：', JSON.stringify(reqt_json, null, 2));

    const resp_data = await post(dest_url, reqt_json);
    const resp_json = utils.parse_resp_data(resp_data);

    winston.info('银联被扫扣款响应报文：', JSON.stringify(resp_json, null, 2));

    if (!utils.verify_resp_sign(resp_json, params.verify_cer)) {
        throw new Error('验证被扫扣款响应报文签名失败');
    }
    winston.info('验证被扫扣款响应报文签名通过.');

    return { reqt: reqt_json, resp: resp_json }
}


module.exports = {
    consume: consume,
}
