/**
 * Copyright (c) 2017 Lucky Byte, Inc.
 */
const debug = require('debug')('unipay');
const crypto = require('crypto');
const moment = require('moment');
const winston = require('winston');
const utils = require('../lib/utils');
const post = require('../lib/post');

/**
 * 银联实名认证，后台交易，银联直接响应结果
 *
 * 参数:
 *     dest_url: 银联实名认证请求地址
 *     params: 一个 JSON 对象，包含下列的 key:
 *      * merno: 商户代码
 *      * order_id: 商户订单号
 *      * cert_sn: 签名证书 Serial Number
 *      * sign_pem: 签名证书，PEM
 *      * acct_no: 卡号/账号
 *      * acct_name: 账户姓名(和 acct_cert_no 必填其一)
 *      * acct_cert_no: 账户身份证号码(和 acct_name 必填其一)
 *      * crypt_cert_sn: 加密证书序号，可选
 *      * crypt_cert: 加密证书，可选
 *      * verify_cer: 验证响应报文签名证书
 *
 * 返回:
 *     JSON，包含 2 个字段:
 *          reqt: 请求报文, JSON
 *          resp: 响应报文, JSON, 通过报文中 respCode 确定交易是否成功
 */
const exchange = async (dest_url, params) => {
    if (!dest_url || !params) {
        throw new Error('银联实名认证请求参数无效，用法 exchange(dest_url, {...})');
    }
    const required_keys = [
        'merno', 'cert_sn', 'sign_pem', 'verify_cer', 'order_id', 'acct_no'
    ]
    for (let i = 0; i < required_keys.length; i++) {
        const key = required_keys[i];
        if (!params[key]) {
            throw new Error(`银联实名认证请求参数缺少[${key}]`);
        }
    }
    if (!params.acct_name && !params.acct_cert_no) {
        throw new Error('实名认证参数 [acct_name] 和 [acct_cert_no] 必填其一');
    }
    let reqt_json = {
        version: '5.0.0',
        encoding: 'UTF-8',
        signMethod: '01',
        txnType: '72',
        txnSubType: '01',
        bizType: '001001',
        channelType: '07',
        accessType: '0',
        txnTime: moment().format('YYYYMMDDHHmmss'),
        certId: params.cert_sn,
        merId: params.merno,
        orderId: params.order_id,
    }
    if (params.version) {
        if (params.version != '5.0.0' && params.version != '5.1.0') {
            throw new Error('参数 version 无效，必须是 5.0.0 或 5.1.0');
        }
        reqt_json.version = params.version;
    }
    if (params.crypt_cert) {
        reqt_json.encryptCertId = params.crypt_cert_sn;
        reqt_json.accNo = crypto.publicEncrypt({
            key: params.crypt_cert,
            padding: crypto.constants.RSA_PKCS1_PADDING
        }, new Buffer(params.acct_no, 'utf-8')).toString('base64');
    } else {
        reqt_json.accNo = params.acct_no;
    }
    reqt_json.customerInfo = utils.build_cust_info(params);
    reqt_json.signature = utils.gen_reqt_sign(reqt_json, params.sign_pem);

    winston.info('银联实名认证请求报文：', JSON.stringify(reqt_json, null, 2));

    const resp_data = await post(dest_url, reqt_json);
    const resp_json = utils.parse_resp_data(resp_data);

    winston.info('银联实名认证响应报文：', JSON.stringify(resp_json, null, 2));

    if (!utils.verify_resp_sign(resp_json, params.verify_cer)) {
        throw new Error('验证实名认证响应报文签名失败');
    }
    debug('验证实名认证响应报文签名通过');

    return { reqt: reqt_json, resp: resp_json }
}


module.exports = {
    exchange: exchange,
}
